'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.walk = walk;
exports.postWalk = postWalk;
exports.preWalk = preWalk;

var _pipe = require('ramda/src/pipe');

var _pipe2 = _interopRequireDefault(_pipe);

var _zipper = require('./zipper');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// @flow


function identity(x) {
    return x;
}

function forEachChild(fn, z) {
    if ((0, _zipper.isLeaf)(z)) return z;
    var _z = z;
    _z = (0, _zipper.whilst)(_zipper.canGoRight, (0, _pipe2.default)(fn, _zipper.right), (0, _zipper.down)(_z));
    return (0, _zipper.up)(fn(_z));
}

/**
 * @namespace walk
 */

/**
 * Walks the data structure in depth-first order,
 * applying inner and outer functions before and after (respectively) each
 * item's subtree is walked.
 *
 * Returns a new data structure from modified items, or the original
 * zipper if the structure wasn't modified.
 *
 * @alias walk.walk
 * @memberof walk
 * @param  {Function} inner - function applied to each item before it's subtree is walked
 * @param  {Function} outer  function applied to each item after it's subtree was walked
 * @param  {Zipper} zipper - A Zipper value to walk
 * @return {Zipper}
 */
function walk(inner, outer, zipper) {
    function innerWalk(z) {
        return (0, _zipper.edit)(outer, forEachChild(innerWalk, (0, _zipper.edit)(inner, z)));
    }
    return innerWalk(zipper);
}

/**
 * Walks the data structure in depth-first order, applying
 * the function after the item's subtree has been walked.
 *
 * Returns a new data structure of modified items, or the original
 * zipper if the structure wasn't modified.
 *
 * @alias walk.postWalk
 * @memberof walk
 * @param  {Function} fn - function applied to each item after it's subtree was walked
 * @param  {Zipper} zipper - A Zipper value to walk
 * @return {Zipper}
 */
function postWalk(fn, zipper) {
    return walk(identity, fn, zipper);
}

/**
 * Walks the data structure in depth-first order, applying
 * the function before the item's subtree has been walked.
 *
 * Returns a new data structure of modified items, or the original
 * zipper if the structure wasn't modified.
 *
 * @alias walk.preWalk
 * @memberof walk
 * @param  {Function} fn - function applied to each item before it's subtree is walked
 * @param  {Zipper} zipper - A Zipper value to walk
 * @return {Zipper}
 */
function preWalk(fn, zipper) {
    return walk(fn, identity, zipper);
}

exports.default = {
    walk: walk,
    postWalk: postWalk,
    preWalk: preWalk
};